require "skirental_api/railtie"

require "faraday"
require "faraday_middleware"
require "faraday-http-cache"

require "symbolize_helper"

# curl -X GET "https://sports-rental.info/srapi/" -H "accept: application/json"

# Fake cache control until we get some from API
class Faraday::OverrideCacheControl < Faraday::Middleware
    def initialize(app, options = {})
        super(app)
        @cache_control = options[:cache_control]
    end
    def call(env)
        response = @app.call(env)
        response.headers["Cache-Control"] = @cache_control
        response
    end
end

module SkirentalApi
    class Client
        using SymbolizeHelper

        def self.service_connection
            config = Rails.application.config.services[:SR_API]
            ocr_service = config[:base_url]
            Faraday.new(:url => ocr_service) do |f|
                f.basic_auth(config[:login], config[:password])
                f.response :logger                                                      # log requests to STDOUT
                f.use Faraday::HttpCache, store: Rails.cache, logger: Rails.logger
                f.adapter  Faraday.default_adapter                                      # make requests with Net::HTTP
                f.use Faraday::OverrideCacheControl, cache_control: 'public, max-age=3600' # Fake cache control data!!
                f.use FaradayMiddleware::ParseJson, content_type: 'application/json'
            end
        end

        def self.sync_booking_state(booking)
            return nil if booking.pre_payment_context[:customers].blank?

            booking_hash = {}
            action = 1
            if booking.booking_code.present?
                booking_hash = self.get_booking(booking.booking_code)
                action = 2
            end

            booking_data = {
                bookings:[
                ],
                orders: [
                ],
                client: {
                    address: {
                        email: booking.pre_payment_context[:email]
                    }
                },
                note: "",
                shop: self.get_mapping[booking.shop_id],
                useInsurance: false, # Not supported
                agent: "",
                action: action
            }

            #byebug TODO: Review merging
            booking_data.merge!(booking_hash)

            # Clear old booking data before rewrite them
            booking_data[:orders] = []

            # Now rewrite complete booking context and ..
            first_run = true
            booking.pre_payment_context[:customers].each do |customer|
                items = []

                customer[:products].each do |product|
                    item = {
                        subitems: [],
                        discnt: product[:price][:discountValue],
                        discntP: product[:price][:discountPercentage],
                        insure: product[:price][:insuranceValue],
                        insureP: product[:price][:insurancePercentage],
                        price: product[:price][:listPrice],
                        total: product[:price][:netPrice],
                        productId: product[:id],
                        product: product[:name],
                        alias: product[:name],
                        modelId: variant_by_product_id(booking, product[:id])[:id],
                        model: variant_by_product_id(booking, product[:id])[:name]
                    }

                    items.push item
                end

                order = {
                    :items => items,
                    :from  => booking.pre_payment_context[:startDate],
                    :until => booking.pre_payment_context[:endDate]
                }

                # TODO: Review with Keller (might be completely wrong)
                # This API seems to expect that one persion is the main customer who
                # has multiple sub-bookings for his relatives or friends.
                if first_run
                    booking_data[:orders].push order

                    # After adding the main customer, we need to switch to the relatives
                    booking_data[:bookings] = []
                    first_run = false
                else
                    new_booking = {
                        orders: []
                    }

                    new_booking[:orders].push order
                    booking_data[:bookings].push new_booking
                end

            end

            # .. send it to the S|R-Api
            response = self.post_booking(self.get_mapping[booking.shop_id], booking_data)

            # Return the created booking code.
            if response.blank? || response.deep_symbolize_keys[:code].blank?
                return nil
            else
                return response.deep_symbolize_keys[:code]
            end
        end

        def self.confirm_booking(booking)
            action = 4

            # voucher füllen
        end

        def self.customer_groups(shop_id, shop_lang)
            all_price_groups = self.price_info(shop_id, shop_lang)

            customer_groups = []
            all_price_groups.each do |price_group|
                customer_goup = {
                    id:             price_group[:iDs].first.to_s,
                    name:           price_group[:name],
                    ageFrom:        price_group[:ageFrom],
                    ageTo:          price_group[:ageTo]
                }
                customer_groups.push customer_goup
            end

            customer_groups.sort_by { |group| group[:ageFrom] }
        end


        def self.valid_booking_range(shop_id, shop_lang)
            dates = self.settings(shop_id, shop_lang)[:dates]
            valid_range = dates[:valid]
            lead_range  = dates[:lead]

            start_date = nil
            end_date = nil
            max_number_of_days = nil
            not_before_min_days = nil
            not_bofore_max_days = nil
            today_ok_time = nil

            if valid_range[:maxLength] > 0
                max_number_of_days = valid_range[:maxLength]
            end

            # die  datumsbereiche sind nur gültig, wenn alle 4 werte gesetzt sind. Sonst ist alles gültig
            if (valid_range[:monthFrom] > 0  && valid_range[:dayFrom] > 0  && valid_range[:monthUntil] > 0 && valid_range[:dayUntil] > 0 )
                start_date = Date.new( Date.current.year, valid_range[:monthFrom], valid_range[:dayFrom] )
                if valid_range[:monthUntil] < valid_range[:monthFrom]
                    end_date = Date.new( Date.current.year + 1, valid_range[:monthUntil], valid_range[:dayUntil] )
                else
                    end_date = Date.new( Date.current.year, valid_range[:monthUntil], valid_range[:dayUntil] )
                end
            else
                start_date = Date.current
                end_date   = Date.current + 120.days # TODO: Review with Keller
            end

            if valid_range[:todayOkTime].present? && valid_range[:todayOkTime] > 0
                today_ok_time = DateTime.now.change(hour: valid_range[:todayOkTime], minute: 0)
            end

            if lead_range[:min] != 0
                not_before_min_days = lead_range[:min]
            end

            if lead_range[:max] != 0
                not_bofore_max_days = lead_range[:max]
            end

            {
                firstValidDay: start_date,
                lastValidDay: end_date,
                lastValidBookingTimeForToday: today_ok_time,
                maxBookingDuration: max_number_of_days,
                minDaysBeforeBooking: not_before_min_days,
                maxDaysBeforeBooking: not_bofore_max_days,
                blockedDays: [], # Not supported
            }
        end

        def self.shop_datum(shop_id, shop_lang)
            shop_data = self.shop(shop_id, shop_lang)
            shop_settings = self.settings(shop_id, shop_lang)
            children_discount = {}
            unless shop_settings[:money][:discount][:kidLimit].nil?
                children_discount = {
                    percentage: shop_settings[:money][:discount][:kidLimit][:value],
                    maxNumberOfChildren: shop_settings[:money][:discount][:kidLimit][:maxkids],
                    maxAgeForChildren: shop_settings[:money][:discount][:kidLimit][:maxage],
                    minNumberOfAdults: shop_settings[:money][:discount][:kidLimit][:minadult]
                }
            end
            early_birds_discounts = []
            unless shop_settings[:money][:discount][:earlyBird].nil?
                shop_settings[:money][:discount][:earlyBird].each do |early_birds_discount|
                    early_birds_discounts.push({
                                                 name: early_birds_discount[:name],
                                                 percentage: early_birds_discounnt[:value],
                                                 daysBeforeBooking: early_birds_discount[:days]
                    })
                end
            end
            payment_providers = []
            shop_settings[:money][:payment][:ipMs].each do |payment_provider|
                payment_providers.push({
                                         id: payment_provider[:id],
                                         type: payment_provider[:type],
                                         discount: payment_provider[:dsc],
                                         show: payment_provider[:show]
                })
            end
            customer_groups = self.customer_groups(shop_id, shop_lang)
            valid_booking_range = self.valid_booking_range(shop_id, shop_lang)
            {
                id: shop_id,
                name: shop_data[:name],
                logoUrl: "http://placehold.it/120x120&text=image1",                                         # TODO: Inhalt von shop_data[:shopInfo][:img] unbrauchbar?!
                title: "#{shop_data[:shopInfo][:title]} #{shop_data[:shopInfo][:title2]}",
                description: "#{shop_data[:shopInfo][:info]} #{shop_data[:shopInfo][:info2]}",
                openingTimes: "#{shop_data[:shopInfo][:open1]} #{shop_data[:shopInfo][:open2]}",
                address: shop_data[:address],
                addressLines: [
                    shop_data[:shopInfo][:addLine1],
                    shop_data[:shopInfo][:addLine2],
                    shop_data[:shopInfo][:addLine3],
                    shop_data[:shopInfo][:addLine4]
                ],
                position: {
                    lat: shop_data[:location][:lat],
                    lng: shop_data[:location][:lng]
                },
                timezone: shop_data[:location][:timezone],
                utcOffset: shop_data[:location][:gmt],
                dateFormat: shop_data[:location][:dateinfo],
                defaultLanguage: shop_settings[:defaults][:deflang],
                currentLanguage: shop_data[:lang],
                currencyCode: shop_settings[:money][:payment][:cur],
                currencySymbol: shop_settings[:money][:cur],
                isPaymentOptional:  shop_settings[:money][:payment][:optional],
                paymentProviders: payment_providers,
                vat: shop_settings[:money][:tax][:value],
                showVat: shop_settings[:money][:tax][:show],
                includeVat: shop_settings[:money][:tax][:include],
                insurance: shop_settings[:money][:insurance][:value],
                showInsurance: shop_settings[:money][:insurance][:show],
                isInsuranceOptional: shop_settings[:money][:insurance][:optional],
                onlineDiscount: shop_settings[:money][:discount][:value],
                childrenDiscount: children_discount,
                earlyBirdDiscounts: early_birds_discounts,
                validBookingRange: valid_booking_range,
                customerGroups: customer_groups
            }
        end

        def self.categories(shop_id, shop_lang)
            categories = self.sr_products(shop_id, shop_lang)
            return_categories = []
            categories.each do |category|
                # Skip if there are no products in this category
                next if category[:products].blank?

                # We don't have an ID. Thus, we use the name as reference. It should be unique
                ret_hash = { id: category[:name], name: category[:name] }
                return_categories.push ret_hash
            end

            return_categories
        end

        def self.packages(shop_id, product_category_id, start_date, duration, customer_group_id, shop_lang)
            age = self.age_for_customer_group(shop_id, shop_lang, customer_group_id)
            categories = self.sr_products(shop_id, shop_lang)
            return_packages = []
            categories.each do |category|
                # Skip if there are no products in this category
                next if category[:products].blank?
                next if category[:packages].blank?

                # This lookup is kind of slow. We might need to improve later, if necessary
                next if category[:name] != product_category_id

                category[:packages].each do |package|
                    # Main Product ist das erste nicht optionale Element in der Liste der items
                    main_product = package[:items].select { |item| item[:optional] == false }.first

                    # Accessories are all products which are marked as optional
                    accessories = package[:items].select { |item| item[:optional] == true }

                    package_hash = {
                        id:             package[:id].to_s,
                        name:           package[:alias].blank? ? package[:name] : package[:alias] ,
                        description:    package[:info],
                        image:          "https://sports-rental.info" + package[:img],
                        mainProduct:    self.convert_item_to_accessory(main_product, start_date, duration, age, shop_id, shop_lang, true),
                        accessories:    self.convert_items_to_accessories(accessories, start_date, duration, age, shop_id, shop_lang, true)
                    }
                    return_packages.push package_hash
                end

            end

            return_packages
        end

        def self.products(shop_id, product_category_id, start_date, duration, customer_group_id, shop_lang)
            age = self.age_for_customer_group(shop_id, shop_lang, customer_group_id)
            categories = self.sr_products(shop_id, shop_lang)
            return_packages = []
            categories.each do |category|
                # Skip if there are no products in this category
                next if category[:products].blank?

                # This lookup is kind of slow. We might need to improve later, if necessary
                next if category[:name] != product_category_id

                return_packages = self.convert_items_to_accessories(category[:products], start_date, duration, age, shop_id, shop_lang, false)
            end

            return_packages
        end


        # Helper Functions
        def self.age_for_customer_group(shop_id, shop_lang, customer_group_id)
            customer_group = self.customer_groups(shop_id, shop_lang).select{ |cg| cg[:id] == customer_group_id }.first

            # Just use the start-Age for selecting the correct price group
            customer_group[:ageFrom]
        end

        # TODO: use map..
        def self.convert_items_to_accessories(items, start_date, duration, age, shop_id, shop_lang, is_package)
            accessories = []
            items.each { |item| accessories.push convert_item_to_accessory(item, start_date, duration, age, shop_id, shop_lang, is_package) }
            accessories
        end

        def self.convert_item_to_accessory(item, start_date, duration, age, shop_id, shop_lang, is_package)
            {
                id:          item[:id].to_s,
                name:        item[:alias].blank? ? item[:name] : item[:alias] ,
                description: item[:details],
                photoUrl:    "https://sports-rental.info" + item[:img],
                variants:    self.convert_models_to_variants( item[:models] ),
                price:       self.get_price_of_item(item[:id], start_date, duration, age, shop_id, shop_lang, is_package)
            }
        end

        def self.get_price_of_item(id, start_date, duration, age, shop_id, shop_lang, is_package)
            item_price = self.price(shop_id, shop_lang, start_date, duration, age, id)
            price_hash = nil

            # Return empty hash if the API has nothing for us!
            return {} if item_price.blank?

            if is_package
                return {} if item_price[:packages].blank?
                price_hash = item_price[:packages].first[:products].first[:prices]
            else
                return {} if item_price[:products].blank?
                price_hash = item_price[:products].first[:prices].first
            end

            {
                listPrice:           price_hash[:basePrc],
                discountPercentage:  price_hash[:discntP],
                discountValue:       price_hash[:discnt],
                insurancePercentage: price_hash[:insuranceP],
                insuranceValue:      price_hash[:insurance],
                netPrice:            price_hash[:value],
                vatPercentage:       price_hash[:taxP],
                vatValue:            price_hash[:tax],
                grossPrice:          price_hash[:value] + price_hash[:tax]
            }
        end

        def self.convert_models_to_variants(models)
            ret_variants = []
            return ret_variants if models.nil?

            models.each do |model|
                variant_hash = { id: model[:vid], name: model[:name] }
                ret_variants.push variant_hash
            end

            ret_variants
        end

        def self.variant_by_product_id(booking, product_id)
            booking.post_payment_context[:product_id] || {}
        end


        # Low-Level requests from SR Backend
        private

        def self.api_version
            response = service_connection.get("/srapi/")
            fail if response.status != 200
            response.body[:version]
        end

        def self.login
            response = service_connection.get("/srapi/login")
            fail if response.status != 200
            response.body
        end

        def self.settings(shop_id, shop_lang)
            response = service_connection.get("/srapi/settings/#{URI.encode(shop_id)}?lang=#{shop_lang}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.sr_products(shop_id, shop_lang)
            response = service_connection.get("/srapi/product/#{URI.encode(shop_id)}?lang=#{shop_lang}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.client(shop_id, shop_lang)
            response = service_connection.get("/srapi/client/#{URI.encode(shop_id)}?lang=#{shop_lang}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.shop(shop_id, shop_lang)
            response = service_connection.get("/srapi/shop/#{URI.encode(shop_id)}?lang=#{shop_lang}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.ticket
            response = service_connection.get("/srapi/ticket")
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.price(shop_id, shop_lang, start, duration, age, id)
            response = service_connection.get("/srapi/price/#{URI.encode(shop_id)}/#{shop_lang}/#{start}/#{duration}/#{age}/#{id}")
            fail if response.status != 200 && response.status != 204
            response.body.deep_symbolize_keys
        end

        def self.price_info(shop_id, shop_lang)
            response = service_connection.get("/srapi/priceinfo/#{URI.encode(shop_id)}?lang=#{shop_lang}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.get_booking(booking_code)
            response = service_connection.get("/srapi/booking/#{booking_code}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.post_booking(numeric_shop_id, booking_data)
            response = service_connection.post do |req|
                req.url("/srapi/booking/#{numeric_shop_id}")
                req.headers['Content-Type'] =   'application/json'
                req.body =                      booking_data.to_json
            end

            Rails.logger.debug "Post-Booking-Response: #{response.body} - Status: #{response.status}"
            byebug if response.status != 200
            fail if response.status != 200

            response.body.deep_symbolize_keys
        end

        def self.get_mapping
            response = service_connection.get( "/srapi/Mapping" )
            fail if response.status != 200
            response.body
        end

    end
end
